@@ -1,3 +1,5 @@
+RSpec::Support.require_rspec_support 'mutex'
+
 module RSpec
   module Mocks
     # A message expectation that only allows concrete return values to be set
@@ -236,6 +238,7 @@ module RSpec
         self.inner_implementation_action = block
         self
       end
+      alias time times
 
       # Expect a message not to be received at all.
       #
@@ -330,7 +333,7 @@ module RSpec
       def ordered(&block)
         if type == :stub
           RSpec.warning(
-            "`allow(...).to receive(..).ordered` is not supported and will" \
+            "`allow(...).to receive(..).ordered` is not supported and will " \
             "have no effect, use `and_return(*ordered_values)` instead."
           )
         end
@@ -357,27 +360,29 @@ module RSpec
       # some collaborators it delegates to for this stuff but for now this was
       # the simplest way to split the public from private stuff to make it
       # easier to publish the docs for the APIs we want published.
+      # rubocop:disable Metrics/ModuleLength
       module ImplementationDetails
         attr_accessor :error_generator, :implementation
         attr_reader :message
         attr_reader :orig_object
         attr_writer :expected_received_count, :expected_from, :argument_list_matcher
-        protected :expected_received_count=, :expected_from=, :error_generator, :error_generator=, :implementation=
+        protected :expected_received_count=, :expected_from=, :error_generator=, :implementation=
 
         # @private
         attr_reader :type
 
-        # rubocop:disable Style/ParameterLists
+        # rubocop:disable Metrics/ParameterLists
         def initialize(error_generator, expectation_ordering, expected_from, method_double,
                        type=:expectation, opts={}, &implementation_block)
           @type = type
           @error_generator = error_generator
-          @error_generator.opts = opts
+          @error_generator.opts = error_generator.opts.merge(opts)
           @expected_from = expected_from
           @method_double = method_double
           @orig_object = @method_double.object
           @message = @method_double.method_name
           @actual_received_count = 0
+          @actual_received_count_write_mutex = Support::Mutex.new
           @expected_received_count = type == :expectation ? 1 : :any
           @argument_list_matcher = ArgumentListMatcher::MATCH_ALL
           @order_group = expectation_ordering
@@ -395,7 +400,7 @@ module RSpec
           @implementation = Implementation.new
           self.inner_implementation_action = implementation_block
         end
-        # rubocop:enable Style/ParameterLists
+        # rubocop:enable Metrics/ParameterLists
 
         def expected_args
           @argument_list_matcher.expected_args
@@ -536,7 +541,9 @@ module RSpec
         end
 
         def increase_actual_received_count!
-          @actual_received_count += 1
+          @actual_received_count_write_mutex.synchronize do
+            @actual_received_count += 1
+          end
         end
 
       private
@@ -567,7 +574,9 @@ module RSpec
             parent_stub.invoke(nil, *args, &block)
           end
         ensure
-          @actual_received_count += increment
+          @actual_received_count_write_mutex.synchronize do
+            @actual_received_count += increment
+          end
         end
 
         def has_been_invoked?
@@ -581,6 +590,7 @@ module RSpec
         end
 
         def set_expected_received_count(relativity, n)
+          raise "`count` is not supported with negative message expectations" if negative?
           @at_least = (relativity == :at_least)
           @at_most  = (relativity == :at_most)
           @exactly  = (relativity == :exactly)
@@ -625,6 +635,7 @@ module RSpec
           nil
         end
       end
+      # rubocop:enable Metrics/ModuleLength
 
       include ImplementationDetails
     end
@@ -650,7 +661,7 @@ module RSpec
             @error_generator.raise_wrong_arity_error(args, block_signature)
           end
 
-          value = @eval_context ? @eval_context.instance_exec(*args, &block) : block.call(*args)
+          value = @eval_context ? @eval_context.instance_exec(*args, &block) : yield(*args)
         end
         value
       end
